% Sequential posterior update for the Saila-Lorda S-R model with an environmental factor, of the form
% log(R_t) =  log(k)+c_SL*(log(S_t)-log(S_k))+c_SL(1-S_t/S_k)+d_SL*E + e, e~N(0,c_t*sigma)
% This version of the fitting code uses an exp-log-subtract trick to prevent
% overflow (in mlikelihoods_new)!
function post_sample = ...
    saila_lorda_2_env_factor_combined_1D_IS_update(prior_sample, post_sample, xt, options, t, T)

obs = xt(1); % log(R)
expl = xt(2); % S
c = xt(3); % c_t in c_t * sigma, not allee parameter c
env1 = xt(4);
env2 = xt(5);

% STEP (0) (set the smoothing parameter h and calculate a)
h = options.solver.h;
a = sqrt(1-h^2); % fixed now, add to options struct

% this should be done in a better way but for testing purposes this is ok
% if prior has finite support, we need to make sure resampled particles fit
LOW = [options.model.upm.prior.hyperparameters.kLow ;
    options.model.upm.prior.hyperparameters.SkLow;
    options.model.upm.prior.hyperparameters.cLow ;
    options.model.upm.prior.hyperparameters.dLow;
    options.model.upm.prior.hyperparameters.fLow;
    options.model.upm.prior.hyperparameters.sigmaLow];
UP = [options.model.upm.prior.hyperparameters.kUp ;
    options.model.upm.prior.hyperparameters.SkUp ;
    options.model.upm.prior.hyperparameters.cUp ;
    options.model.upm.prior.hyperparameters.dUp ;
    options.model.upm.prior.hyperparameters.fUp ;
    options.model.upm.prior.hyperparameters.sigmaUp];

N = options.solver.n_particles;

for i = 1:length(post_sample)
    ks = post_sample{i}(1,:);
    Sks = post_sample{i}(2,:);
    cs = post_sample{i}(3,:);
    ds = post_sample{i}(4,:);
    fs = post_sample{i}(5,:);
    sigmas = post_sample{i}(6,:);
    weights = post_sample{i}(7,:);
    
    % STEP (10) (calculate the number of effective particles)
    N_eff = 1/sum((sort(weights)).^2);
    
    % print the number of effective particles before update
    disp(' ')
    disp(['t = ' num2str(t) '/' num2str(T)...
          ', r = ' num2str(i) ', N_eff before update = ' num2str(N_eff)])
    
    switch lower(options.model.upm.likelihood.type)
        case 'normal'
            
            N_eff = 0; % set to zero to initialize while loop
            
            while N_eff < 0.5*N
                % STEP (1) (take the parameter vectors from the previous 
                %           time step)
                thetas = zeros(4,N);
                thetas(1,:) = ks;
                thetas(2,:) = Sks;
                thetas(3,:) = cs;
                thetas(4,:) = ds;
                thetas(5,:) = fs;
                thetas(6,:) = sigmas;

                % STEP (2) (calculate the sample mean and covariance matrix
                % from the particle locations)                
                thetasbar = mean(thetas,2);
                V = cov(thetas');
                
                % STEP (3) (calculate the means of the kernel locations 
                ms = a*thetas+repmat((1-a)*thetasbar,1,N);
                
                thetas_new = zeros(size(thetas,1),N);
                
                % STEP (4) AND (5) (calculate weights proportional 
                %                   to the product of the likelihood function evaluated 
                %                   at the kernel means and the weights at the previous time step)
                % log(R_t) =  log(k)+c_SL*(log(S_t)-log(S_k))+c_SL(1-S_t/S_k)+e
                %
                %mlikelihoods in log-scale
                %mus0=ms(1,:)+exp(ms(3,:)).*(log(expl)-ms(2,:))+exp(ms(3,:)).*(1-expl./exp(ms(2,:)));
                %sigmas0=c*exp(ms(5,:));
                mlikelihoods_log = -log(sqrt(2*pi*(...
                                    (c*ms(6,:))... %sigmas0
                                    .^2)))-1./(2*(...
                                    (c*ms(6,:))... %sigmas0
                                    .^2)).*((obs-...
                                    (log(ms(1,:))+ms(3,:).*(log(expl)-log(ms(2,:)))+ms(3,:).*(1-expl./ms(2,:)) + env1.*ms(4,:) + env2.*ms(5,:) ...
                                    -(ms(6,:).^2)./2)... %mus0
                                    ).^2); 
                %log(weights.*mlikelihoods)
                %mweights in log-scale
                mweights = log(weights)+mlikelihoods_log;
                %subtract a constant
                mweights=mweights-max(mweights); 
                mweights=exp(mweights);%=mweights*A with A=exp(-max(mweights_log))

                % normalize mweights, A cancels out
                mweights = mweights/sum(sort(mweights));
                
                % STEP (6) (sample N indices from a multinomial distribution according to the weights)
                minds = randsample(1:N,N,true,mweights);
                %randsample uses a vector of nonnegative weights (mweights) of the same length as the 
                %vector 1:N, to determine the probability that a value is
                %selected from this vector
                
                % STEP (7) (draw new particle locations in the log-space from a multivariate normal
                % distribution (and check that they belong to the support of the prior distributions,
                % otherwise draw again))
                for j = 1:N
                    proposal = mvnrnd(ms(:,minds(j)),h^2*V);
                    proposal = proposal(:);
                    if all(proposal >= LOW & proposal <= UP)
                        thetas_new(:,j) = proposal;
                    else
                        while any(proposal < LOW | proposal > UP)
                            proposal = mvnrnd(ms(:,minds(j)),h^2*V);
                            proposal = proposal(:);
                        end
                        thetas_new(:,j) = proposal;
                    end
                end
                ks_new = thetas_new(1,:);
                Sks_new = thetas_new(2,:);
                cs_new = thetas_new(3,:);
                ds_new = thetas_new(4,:);
                fs_new = thetas_new(5,:);
                sigmas_new = thetas_new(6,:);
                
                % STEP (8) (get the new particle locations)
                
                ks = ks_new;
                Sks = Sks_new;
                cs = cs_new;
                ds = ds_new;
                fs = fs_new;
                sigmas = sigmas_new;
                
                % STEP (9) (calculate the new particle weights proportional to the ratio
                % of the likelihoods evaluated at the kernel means and at the new particle locations)
                % 
                % log(R_t) =  log(k)+c_SL*(log(S_t)-log(S_k))+c_SL(1-S_t/S_k)+e
                %mus1=ms(1,minds)+exp(ms(3,minds)).*(log(expl)-ms(2,minds))+exp(ms(3,minds)).*(1-expl./exp(ms(2,minds)));
                %sigmas1=c*exp(ms(5,minds));
                mlikelihoods_new_log = -log(sqrt(2*pi*(...
                                        (c*ms(6,minds))...
                                        .^2)))-1./(2*(...
                                        (c*ms(6,minds))...
                                        .^2)).*((obs-...
                                        (log(ms(1,minds))+ms(3,minds).*(log(expl)-log(ms(2,minds)))+ms(3,minds).*(1-expl./ms(2,minds)) + env1.*ms(4,minds) + env2.*ms(5,minds) ...
                                        -(ms(6,minds).^2)./2)...
                                        ).^2); 
                
                % likelihoods evaluated at the new particle locations?
                %mus2=logks_new+cs_new.*(log(expl)-logSks_new)+cs_new.*(1-expl./Sks_new);
                %sigmas2=c*sigmas_new; %sigmas
                likelihoods_log =  -log(sqrt(2*pi*(...
                                    (c*sigmas_new)... %sigmas2
                                    .^2)))-1./(2*((...
                                    c*sigmas_new)... %sigmas2
                                    .^2)).*((obs-...
                                    (log(ks_new)+cs_new.*(log(expl)-log(Sks_new))+cs_new.*(1-expl./Sks_new)  + env1.*ds_new + env2.*fs_new...
                                    -(sigmas_new.^2)./2)... % mus2
                                    ).^2); 

                weights_new_log = likelihoods_log-mlikelihoods_new_log;%=log(likelihoods./mlikelihoods_new)
                weights_new_log=weights_new_log-max(weights_new_log); %subtract a constant

                weights_new=exp(weights_new_log); %=weights_new*B with B=exp(-max(weights_new_log))
                sorted_weights_new=sort(weights_new);
                const=sum(sorted_weights_new);
                
                N_eff = 1/sum(sorted_weights_new.^2)*(const^2);%by multiplying with const^2 we obtain 
                %N_eff = 1/sum(weights_new.^2);
                disp(['t = ' num2str(t) '/' num2str(T)...
                      ', r = ' num2str(i) ', N_eff after update = ' num2str(N_eff)])
                
            end
            
        case 'student-t'
            error('student-t error not implemented yet')
        case 'uniform'
            error('uniform error not implemented yet')
        otherwise
            error('Undefined error type')
    end
    
    post_sample{i}(1,:) = ks_new;
    post_sample{i}(2,:) = Sks_new;
    post_sample{i}(3,:) = cs_new;
    post_sample{i}(4,:) = ds_new;
    post_sample{i}(5,:) = fs_new;
    post_sample{i}(6,:) = sigmas_new;
    %normalize weights
    post_sample{i}(7,:)=weights_new./const;
    
end

post_sample = [prior_sample ; post_sample];

