% Sequential posterior update for the Sigmoid Beverton-Holt S-R model with an environmental factor, of the form
% log(R) =  log(Rmax)-log((1-q)/q*(Sq/S)^c+1)+ d*E + e, e~N(0,c_t*sigma)
% This version of the fitting code uses an exp-log-subtract trick to prevent
% overflow (in mlikelihoods_new)!
function post_sample = ...
    sigmoid_beverton_holt_env_factor_combined_1D_IS_update(prior_sample, post_sample, xt, options, t, T)

obs = xt(1); % log(R)
expl = xt(2); % S
c = xt(3); % c_t in c_t * sigma, not allee parameter c
env1 = xt(4);
env2 = xt(5);
q = options.model.upm.prior.q;
p = (1-q)/q;

% STEP (0) (set the smoothing parameter h and calculate a)
h = options.solver.h;
a = sqrt(1-h^2); % fixed now, add to options struct

% this should be done in a better way but for testing purposes this is ok
% if prior has finite support, we need to make sure resampled particles fit
LOW = [options.model.upm.prior.hyperparameters.RmaxLow ;
    options.model.upm.prior.hyperparameters.SqLow ;
    options.model.upm.prior.hyperparameters.cLow ;
    options.model.upm.prior.hyperparameters.dLow ;
    options.model.upm.prior.hyperparameters.fLow ;
    options.model.upm.prior.hyperparameters.sigmaLow];
UP = [options.model.upm.prior.hyperparameters.RmaxUp ;
    options.model.upm.prior.hyperparameters.SqUp ;
    options.model.upm.prior.hyperparameters.cUp ;
    options.model.upm.prior.hyperparameters.dUp ;
    options.model.upm.prior.hyperparameters.fUp ;
    options.model.upm.prior.hyperparameters.sigmaUp];

N = options.solver.n_particles;

for i = 1:length(post_sample)
    Rmaxs = post_sample{i}(1,:);
    Sqs = post_sample{i}(2,:);
    cs = post_sample{i}(3,:);
    ds = post_sample{i}(4,:);
    fs = post_sample{i}(5,:);
    sigmas = post_sample{i}(6,:);
    weights = post_sample{i}(7,:);
    
    % STEP (10) (calculate the number of effective particles)
    N_eff = 1/sum((sort(weights)).^2);
    
    % print the number of effective particles before update
    disp(' ')
    disp(['t = ' num2str(t) '/' num2str(T) ', r = ' num2str(i) ', N_eff before update = ' num2str(N_eff)])
    
    switch lower(options.model.upm.likelihood.type)
        case 'normal'
            
            N_eff = 0; % set to zero to initialize while loop
            
            while N_eff < 0.5*N
                % STEP (1) (take the parameter vectors from the previous 
                %           time step)
                %thetas = [log(Rmaxs) ; log(Sqs) ; log(cs) ; log(ds); log(sigmas)];
                thetas = [Rmaxs ; Sqs ; cs ; ds; fs; sigmas];
                
                % STEP (2) (calculate the sample mean and covariance matrix
                % from the  particle locations)
                thetasbar = mean(thetas,2);
                V = cov(thetas');
                %end
                
                % STEP (3) (calculate the means of the kernel locations 
                ms = a*thetas+repmat((1-a)*thetasbar,1,N);
                
                thetas_new = zeros(size(thetas,1),N);
                
                % STEP (4) AND (5) (calculate weights proportional 
                %                   to the product of the likelihood function evaluated 
                %                   at the kernel means and the weights at the previous time step)                
                %log(R) =  log(Rmax)-log((1-q)/q*(Sq/S)^c+1)+e 
                %
                %mlikelihoods in log-scale
                %mus0=ms(1,:)-log(p*(exp(ms(2,:))./expl).^exp(ms(3,:))+1);
                %sigmas0=c*exp(ms(5,:));
                mlikelihoods_log = -log(sqrt(2*pi*(...
                                    (c*ms(6,:))...
                                    .^2)))-1./(2*(...
                                    (c*ms(6,:))...
                                    .^2)).*((obs-...
                                    (log(ms(1,:))-log(p*(ms(2,:)./expl).^ms(3,:)+1) + env1.*ms(4,:) + env2.*ms(5,:)...
                                    -(ms(6,:).^2)./2)).^2);  
                %log(weights.*mlikelihoods)
                %mweights in log-scale
                mweights = log(weights)+mlikelihoods_log;
                %subtract a constant
                mweights = mweights-max(mweights); 
                mweights=exp(mweights);%=mweights*A with A=exp(-max(mweights_log))

                % normalize mweights, A cancels out
                mweights = mweights/sum(sort(mweights));
                
                % STEP (6) (sample N indices from a multinomial distribution according to the weights)
                minds = randsample(1:N,N,true,mweights);
                
                % STEP (7) (draw new particle locations from a multivariate normal
                % distribution (and check that they belong to the support of the prior distributions,
                % otherwise draw again))
                for j = 1:N
                    proposal = mvnrnd(ms(:,minds(j)),h^2*V);
                    proposal = proposal(:);
                    if all(proposal >= LOW & proposal <= UP)
                        thetas_new(:,j) = proposal;
                    else
                        while any(proposal < LOW | proposal > UP)
                            proposal = mvnrnd(ms(:,minds(j)),h^2*V);
                            proposal = proposal(:);
                        end
                        thetas_new(:,j) = proposal;
                    end                    
                end
                Rmaxs_new = thetas_new(1,:);
                Sqs_new = thetas_new(2,:);
                cs_new = thetas_new(3,:);
                ds_new = thetas_new(4,:);
                fs_new = thetas_new(5,:);
                sigmas_new = thetas_new(6,:);
                
                % STEP (8) (get the new particle locations)
                
                Rmaxs = Rmaxs_new;
                Sqs = Sqs_new;
                cs = cs_new;
                ds = ds_new;
                fs = fs_new;
                sigmas = sigmas_new;
                
                % STEP (9) (calculate the new particle weights proportional to the ratio
                % of the likelihoods evaluated at the kernel means and at the new particle locations)
                %
                % log(R) =  log(Rmax)-log((1-q)/q*(Sq/S)^c+1)+e
                %mus1=ms(1,minds)-log(p*(exp(ms(2,minds))./expl).^exp(ms(3,minds))+1);
                %sigmas1=c.*exp(ms(4,minds));
                mlikelihoods_new_log = -log(sqrt(2*pi*(...
                                            (c.*ms(6,minds))...
                                            .^2)))-1./(2*(...
                                            (c.*ms(6,minds))...
                                            .^2)).*((obs-...
                                            (log(ms(1,minds))-log(p*(ms(2,minds)./expl).^ms(3,minds)+1) + env1.*ms(4,minds) + env2.*ms(5,minds)...
                                            -(ms(6,minds).^2)./2)...
                                            ).^2); 
                
                %mus2=logRmaxs_new-log(p*(Sqs_new/expl).^cs_new+1);
                likelihoods_log =  -log(sqrt(2*pi*(sigmas_new.^2)))-1./(2*(sigmas_new.^2)).*((obs-...
                                    (log(Rmaxs_new)-log(p*(Sqs_new/expl).^cs_new+1) + env1.*ds_new + env2.*fs_new...
                                    -(sigmas_new.^2)./2)...
                                    ).^2); 
                
                %weights_new in log-scale
                weights_new = likelihoods_log-mlikelihoods_new_log;%=log(likelihoods./mlikelihoods_new)
                weights_new=weights_new-max(weights_new); %subtract a constant
                
                weights_new=exp(weights_new); %weights_new_multiplied_by_constant=weights_new*B with B=exp(-max(weights_new_log))
                sorted_weights_new=sort(weights_new);
                const=sum(sorted_weights_new);
                
                N_eff = 1/sum(sorted_weights_new.^2)*(const^2);%by multiplying with const^2 we obtain 
                %N_eff = 1/sum(weights_new.^2);
                disp(['t = ' num2str(t) '/' num2str(T) ', r = ' num2str(i) ', N_eff after update = ' num2str(N_eff)])
                
            end
            
        case 'student-t'
            error('student-t error not implemented yet')
        case 'uniform'
            error('uniform error not implemented yet')
        otherwise
            error('Undefined error type')
    end
    
    post_sample{i}(1,:) = Rmaxs_new;
    post_sample{i}(2,:) = Sqs_new;
    post_sample{i}(3,:) = cs_new;
    post_sample{i}(4,:) = ds_new;
    post_sample{i}(5,:) = fs_new;
    post_sample{i}(6,:) = sigmas_new;
    %normalize weights
    post_sample{i}(7,:) = weights_new./const;
    
end

post_sample = [prior_sample ; post_sample];

