% Tommi Per?l? (tommi.perala@helsinki.fi)
% Modularized version of Bayesian online change point detection
% Adapted from work of
% - Ryan Turner (rt324@cam.ac.uk)
% - Yunus Saatci (ys267@cam.ac.uk)
% - Ryan Adams (rpa23@cam.ac.uk)
%
% + Implements smoothing and prediction
% + Implements point mass approximation so that we can use arbitrary models
% and distributions
%
% Input:
% X  T x 1 variable containing the time series.
% options.UPM_type  function handle to update the posteriors and evaluate the posterior
% predictive density.
% options.prior_params 1 x param_count parameters of the priors on new distributions drawn at
% each time point.
% hazard_f function handle calculating the hazard of a change point.
% options.hazard_type 1 x hazard_param_count parameters to the hazard function
%
% Outputs:
% R T + 1 x T + 1 Object giving the run-length probabilities, R(r, t) =
% P(runlength_t-1 = r-1|X_1:t-1). [P]
% lml log marginal likelihood of the data, X(1:end), under the model
% = P(X_1:T), integrating out all the runlengths. [log P]
%
function BCPD_OUTPUT = ...
    bocpd_smoother_predicter_IS_memory_efficient(X, options)

%
%assert(isKosher(X));
[T,D] = size(X); % Number of time point observed. 1 x 1. [s]
l_r0 = length(options.model.runlength0);

hazard_f = str2func([options.model.hazard.type, '_h']);
model_update_f = str2func([options.model.upm.type, '_update']);
model_predict_f = str2func([options.model.upm.type, '_predict']);
model_init_f = str2func([options.model.upm.type, '_init']);

theta_h = options.model.hazard.hyperparameters;

% R(r, t) = P(runlength_t-1 = r-1|X_1:t-1).
R = zeros(T + 1 + l_r0 - 1, T + 1); % pre-load the run length distribution. [P]
S = zeros(T + l_r0 - 1, T);
RP = zeros(T + 2 + l_r0 - 1, T + 1);
SP = zeros(T + 1 + l_r0 - 1, T + 1);
RSs = cell(T, 1);

% At time t = 1, we actually have complete knowledge about the run
% length.  It is definitely zero.  See the paper for other possible
% boundary conditions. This assumes there was surely a change point right before
% the first data point not at the first data point.
% Implements step 1, alg 1, of [RPA].
% => P(runglenth_0 = 0|nothing) = 1
%R(1, 1) = 1; % 1 x 1. [P]

R(1:l_r0, 1) = options.model.runlength0;

% Calculate one-step predictions
HP = feval(hazard_f, (1:l_r0)', theta_h); % t+1 x 1. [P]
RP(2:l_r0 + 1, 1) = exp(log(R(1:l_r0, 1)) + log((1 - HP))); % t x 1. [P]
RP(1, 1) = sum(exp(log(R(1:l_r0, 1)) + log(HP))); % 1 x 1. [P]
SP(1, 1) = 1;

Z = zeros(T, 1); % The evidence at each time step => Z(t) = P(X_t|X_1:t-1). [P]
% Find parameters of p(X_1|nothing). 1 x param_count. units depend on options.UPM_type.

% initialize the posterior with the prior.

for i = 1:l_r0
    post_sample{i,1} = feval(model_init_f, T + 1, D, options);
end

for t = 1:T
    % Implictly Implements step 2, alg 1, of [RPA]: oberserve new datum, simply by
    % incrementing the loop index.
    
    % Evaluate the hazard function for this interval.
    % H(r) = P(runlength_t = 0|runlength_t-1 = r-1)
    % Pre-computed the hazard in preperation for steps 4 & 5, alg 1, of [RPA]
    H = feval(hazard_f, (1:(t+l_r0-1))', theta_h); % t x 1. [P]
    
    % Evaluate the predictive distribution for the new datum under each of
    % the parameters.  This is the standard thing from Bayesian inference.
    % Implements step 3, alg 1, of [RPA].
    % predprobs(r) = p(X(t)|X(1:t-1), runlength_t-1 = r-1). t x 1. [P]
    if any(isnan(X(t,:)))
        R(1:t+1+l_r0-1,t+1) = RP(1:t+1+l_r0-1,t);
        post_sample = [post_sample{1} ; post_sample];
        % MT: Update S, too. Do not compute evidence if there are missing
        % data.
        S(1:(t+l_r0-1),t)=SP(1:(t+l_r0-1),t);
        Z(t) =  NaN;
    else
        predprobs = feval(model_predict_f, post_sample, X(t, :), options);
        
        % Evaluate the growth probabilities - shift the probabilities up and to
        % the right, scaled by the hazard function and the predictive
        % probabilities.
        % Implements step 4, alg 1, of [RPA].
        % Assigning P(runlength_t = 1|X_1:t) to P(runlength_t = t|X_1:t):
        % P(runlength_t = r|X_1:t) propto P(runlength_t-1 = r-1|X_1:t-1) *
        % p(X_t|X_1:t-1,runlength_t-1 = r-1) * P(runlength_t = r|runlength_t-1 = r-1).
        %R(2:t + 1, t + 1) = R(1:t, t) .* predprobs .* (1 - H); % t x 1. [P]
        R(2:(t+l_r0-1) + 1, t + 1) = exp(log(R(1:(t+l_r0-1), t)) + log(predprobs) + log(1 - H)); % t x 1. [P]
        
        % Evaluate the probability that there *was* a changepoint and we're
        % accumulating the mass back down at r = 0.
        % Implements step 5, alg 1, of [RPA].
        % Assigning P(runlength_t = 0|X_1:t)
        % P(runlength_t = 0|X_1:t) propto sum_r=0^t-1 P(runlength_t-1 = r|X_1:t-1) *
        % p(X_t|X_1:t-1, runlength_t-1 = r) * P(runlength_t = 0|runlength_t-1 = r).
        %R(1, t + 1) = sum(R(1:t, t) .* predprobs .* H); % 1 x 1. [P]
        R(1, t + 1) = sum(exp(log(R(1:(t+l_r0-1), t)) + log(predprobs) + log(H))); % 1 x 1. [P]
        
        % Renormalize the run length probabilities for improved numerical
        % stability.
        % note that unlike in [RPA] which keeps track of P(r_t, X_1:t), we keep track
        % of P(r_t|X_1:t) => unnormalized R(i, t+1) = P(runlength_t = i-1|X_1:t) *
        % P(X_t|X_1:t-1) => normalization const Z(t) = P(X_t|X_1:t-1).
        % Sort of Implements step 6, alg 1, of [RPA].
        % Could sum R(:, t+1), but we only sum R(1:t+1,t+1) to avoid wasting time
        % summing zeros.
        Z(t) =  sum(R(1:(t+l_r0-1) + 1, t + 1)); % 1 x 1. [P]
        % Implements step 7, alg 1, of [RPA].
        % After normalization, R(i, t+1) = P(runlength_t = i-1|X_1:t).
        %R(1:t + 1, t + 1) = R(1:t + 1, t + 1) ./  Z(t); % (T + 1) x 1. [P]
        R(1:(t+l_r0-1) + 1, t + 1) = exp(log(R(1:(t+l_r0-1) + 1, t + 1)) - log(Z(t))); % (T + 1) x 1. [P]
        
        % quick hack to ensure no zeros
        if any(R(1:(t+l_r0-1)+1,t+1) == 0)
            R(R(1:(t+l_r0-1)+1,t+1) == 0,t+1) = eps;
            R(1:(t+l_r0-1) + 1, t + 1) = R(1:(t+l_r0-1) + 1, t + 1)/sum(R(1:(t+l_r0-1) + 1, t + 1));
        end
        
        %S(1:t, t) = R(1:t, t) .* predprobs;
        %S(:, t) = S(:, t) ./ sum(S(:, t));
        %S(1:t, t) = R(1:t, t) .* predprobs / Z(t);
        S(1:(t+l_r0-1), t) = exp(log(R(1:(t+l_r0-1), t)) + log(predprobs) - log(Z(t)));
                
        % post_params(r, i) = current state of hyper-parameter i given runlength = r-1
        % Implements step 8, alg 1, of [RPA].
        % (t + 1) x param_count. units depend on options.UPM_type.
        prior_sample = feval(model_init_f, T + 1, D, options);
        post_sample = feval(model_update_f, prior_sample, post_sample, X(t, :), options, t, T);
    end
    
    current_seed = rng;
    
    %%% MARGINAL POSTERIOR DISTRIBUTIONS FOR MODEL PARAMETERS
    %%% sample the posterior distribution of model parameters online (the mixture distribution consisting of 
    %%% run-specific marginal posterior distributions weighted by run length probabilities) % MT
    POSTERIOR.subsamplesOfParameterPosteriorsWithMixture{t} = ...
            drawPosteriorSampleFullOnline(S,post_sample,t,300);
    % subsamples of run-specific marginal posterior distributions of the model parameters
    POSTERIOR.runSpecificSubsamplesOfParameterPosteriors{t}=drawPosteriorSampleRunSpecificOnline(post_sample,t,10000); %give the number of samples
    
    rng(current_seed)
    
    % Calculate one-step predictions
    HP = feval(hazard_f, (1:(t+l_r0-1)+1)', theta_h); % t+1 x 1. [P]
    %RP(2:t + 2, t + 1) = R(1:t + 1, t + 1) .* (1 - HP); % t x 1. [P]
    RP(2:(t+l_r0-1) + 2, t + 1) = exp(log(R(1:(t+l_r0-1) + 1, t + 1)) + log((1 - HP))); % t x 1. [P]
    %RP(1, t + 1) = sum(R(1:t + 1, t + 1) .* HP); % 1 x 1. [P]
    RP(1, t + 1) = sum(exp(log(R(1:(t+l_r0-1) + 1, t + 1)) + log(HP))); % 1 x 1. [P]
    %RP(2:t + 2, t + 1) = R(1:t + 1, t + 1) .* (1 - H(1)); % t x 1. [P]
    %RP(1, t + 1) = sum(R(1:t + 1, t + 1) .* H(1)); % 1 x 1. [P]
    SP(1:t+1, t+1) = RP(2:t+2, t+1) / sum(RP(2:t+2, t+1));
    
    % Smoothed run-length distribution after each step
    %RSs{t} = zeros((t+l_r0-1)+1,t+1);
    %RSs{t}(1:(t+l_r0-1)+1,t+1) = R(1:(t+l_r0-1)+1,t+1);
    RSs{t} = zeros(t+l_r0-1, t);
    RSs{t}(1:(t+l_r0-1), t) = S(1:(t+l_r0-1), t);
    
    % for k = t:-1:1
    %     RSs{t}(1:(k+l_r0-1),k) = exp(...
    %         log(R(1:(k+l_r0-1),k))...
    %         +log(...
    %         exp(log(H(1:(k+l_r0-1)))+log(RSs{t}(1,k+1))-log(RP(1,k)))...
    %         +exp(...
    %         log((1-H(1:(k+l_r0-1))))...
    %         +log(RSs{t}(2:(k+l_r0-1)+1,k+1))...
    %         -log(RP(2:(k+l_r0-1)+1,k))...
    %         )));
    % end
    for k = t-1:-1:1
        RSs{t}(1:(k+l_r0-1), k) = ...
        exp( ...
            log( S(1:(k+l_r0-1), k) ) + ...
            log( ...
                exp( ...
                    log( H(1:(k+l_r0-1)) ) + ...
                    log( RSs{t}(1, k+1) ) - ...
                    log( SP(1, k+1) ) ...
                    ) + ...
                exp( ...
                    log( 1 - H(1:(k+l_r0-1)) ) + ...
                    log( RSs{t}(2:(k+l_r0-1)+1, k+1) ) - ...
                    log( SP(2:(k+l_r0-1)+1, k+1) ) ...
                ) ...
            ) ...
        );
    end
    %S(1:t,1:t)
    %SP(1:t+1,1:t+1)
    %RSs{t}
    
end

% Get the log marginal likelihood of the data, X(1:end), under the model
% = P(X_1:T), integrating out all the runlengths. 1 x 1. [log P]
nlml = -sum(log(Z));

BCPD_OUTPUT = struct('R',R,'S',S,'nlml',nlml,'Z',Z,...
    'POSTERIOR',{POSTERIOR},...
    'RP',RP,'RSs',{RSs});